{  PPC C2 Flash Loader.spin

          **************************************
               Flash Silabs C8051F micros
                   via C2 interface
          **************************************
                    Peter Jakacki
                      2009-03-16                
          **************************************

PPC chips based upon Silabs C8051F3xx can be programmed over
the C2 interface using 2 wires. In a Propeller system one of the
C2 signals is shared with the I2C bus SDA while the other is a
dedicated Reset/Clock (C2CK) combo.

The C2CLK could be shared with another pin via a jumper if resources
are scarce.

This is a run-once program which once it has programmed the PPC no
longer needs to be resident so use F10 from the Spin tool or
run from RAM.

}
''****************************************************************    

CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000
  clockfreq = ((_CLKMODE - XTAL1) >> 6) * _XINFREQ

   _1mS  = clockfreq /     1_000 'Divisor for  1 mS

'******************* SILABS C2 INTERFACE*********************

  C2CK2         = 19            ' user specific PPC C2CK lines
  C2CK1         = 23
  C2CK0         = 25
  C2D           = 29            ' uses same pin as SDA on Prop
  
' coms  
  txdcoms        = 30
  rxdcoms        = 31 


' C2 constants
  DEVID         = 0
  FPCTL         = 2
  FPDAT         = $B4
  inbusy        = 2
  outready      = 1

  
'
VAR
  word  cold,c2cmd,c2dat
  
      
DAT

VER     byte " HARDWARE: PROPELLER",$0d,$0a
        byte " FIRMWARE: C2 Flash Loader V1.0",$0d,$0a,0

' Image of USER PPC firmware file 
'
PPC     long  @endPPC-@PPC      ' length field
        file  "PPC.BIN"         ' user PPC file for C8051F33x
endPPC

CHIPS   byte  $F2,$F4,$F0       ' list of user I2C address for each chip
endCHIPS

C2CK    byte  C2CK0,C2CK1,C2CK2    



PUB Main | i
  if cold == $a55a                                      ' only want to run once - in case this has been loaded into EEPROM
    REBOOT
  cognew(@C2CODE,@c2cmd)
  i~
  repeat endCHIPS-CHIPS                                 ' Set and program each chip from the CHIP list
     C2SET(byte[@C2CK][i])                              ' Use the chips C2CK/Reset line
     byte[@PPC][$84] := byte[@CHIPS][i++]               ' Set I2C address of chip within firmware image
     C2RESET                    
     \PROGRAM(@PPC)                                     ' Flash the firmware into the chip
  cold := $A55A      
  REBOOT
      
  
pub PROGRAM(fileptr) | faddr,flen,flen2
  C2RESET
  if RDATA <> $0A                                       ' C8051F33x device ID?
    abort string(" NO DEVICE")
  C2RESET
  C2INIT
  C2DeviceErase                        
  VerifyErased
  faddr~
  flen := long[fileptr]                                 
  fileptr += 4
  repeat (flen>>8)+1                                    ' Program in blocks of 256 bytes or less
    flen2 := flen
    flen2 <#= 256
    ProgramBlock(fileptr,faddr,flen2)
    faddr += 256
    flen -= 256
  C2RESET

pub ProgramBlock(src,addr,len) | i
  WADDRESS(FPDAT)                                       '
  WDATA(07)                                             ' WRITE BLOCK COMMAND
  if RDATAW <> $0d                                      ' command accepted?                                                
    abort string(" REJECT COMMAND")
  WDATA(addr>>8)                                        ' ADDRH
  WDATA(addr&$ff)                                       ' ADDRL
  WDATA(len)                                            ' LEN
  if RDATAW <> $0d
    abort string(" REJECT ADDRESS")
  repeat i from addr to addr+len-1
    WDATA(byte[i][src])  

pub PollInBusy
  repeat while RSTATUS&inbusy
  
pub PollOutReady
  repeat until RSTATUS&outready

pub VerifyErased

pub C2DeviceErase
  WADDRESS(FPDAT)
  WDATA(03)
  if RDATAW <> $0d                                                                                      
    abort string(" REJECT ERASE,")
  WDATA($DE)
  WDATA($AD)
  WDATA($A5)
  if RDATAW <> $0d                                                                                      
    abort string(" ERASE FAIL,")

pub command(cmd,data)
  c2dat := data
  c2cmd := cmd
  repeat until c2cmd == 0
  result := c2dat
  
pub C2RESET
  command("Z",0)                  ' reset

pub C2SET(data)
  command("=",data)

pub RDATA                                               ' Read data from C2 register
  result := command("@",0)

pub RDATAW                                              ' 
  PollOutReady
  result := RDATA

pub WDATA(data)
  command("!",data)
  PollInBusy

pub WDATA1(data)
  command("!",data)


pub WADDRESS(data)
  command(":",data)

pub RSTATUS
  result := command("?",0)

pub C2INIT
  WADDRESS(FPCTL)
  WDATA(2)
  WDATA(1)
  ms(20)

PUB ms(Period)                                    '1 mS 
    waitcnt(_1mS * Period + cnt)


DAT
'
' C2 cog required as C2 clock pulses cannot be low longer than 5us which is not possible in Spin
'
              org   0
C2CODE
              mov       par2,par                ' pointer to command word in hub memory
              add       par2,#2                 ' pointer to data word
              or        outa,_c2d
              or        dira,_c2d
CLEAR
              wrword    zero,par2
C2CMDS        wrword    zero,par                  ' clear command
C2CMDLP
              rdword    r0,par                  ' read command
              cmp       r0,#0 wz
        if_z  jmp       #C2CMDLP
              cmp       r0,#"Z"  wz             ' Reset command?
        if_z  jmp       #Reset
              cmp       r0,#"@" wz              ' data ?
        if_z  jmp       #ReadData
              cmp       r0,#":" wz              ' Write Address?
        if_z  jmp       #WriteAddress
              cmp       r0,#"!" wz
        if_z  jmp       #WriteData
              cmp       r0,#"?" wz              ' read status
        if_z  jmp       #ReadAddress
              cmp       r0,#"=" wz              ' set c2ck
        if_z  jmp       #SetC2
              jmp       #C2CMDS

SetC2
              rdword    r0,par2
              mov       _c2ck,#1
              shl       _c2ck,r0
              jmp       #C2CMDS
                             
WriteAddress
              rdword    r0,par2    
              call      #WA
              jmp       #C2CMDS

ReadAddress
              call      #RA
              jmp       #WRPAR 

WriteData
              rdword    r0,par2    
              call      #WD
              jmp       #C2CMDS
              
ReadData
              call      #RD
WRPAR
              wrword    r0,par2    
              jmp       #C2CMDS 
              

'************************* C2 PRIMITIVES *********************        

Reset         call      #RS
              jmp       #C2CMDS

RS
              andn      outa,_c2ck
              or        dira,_c2ck
              mov       r0,#250
rstdly        nop
              djnz      r0,#rstdly
              or        outa,_c2ck
RS_ret        ret

INS
   '' INS field (00b, LSB first)
              shr       r2,#1 wc
              muxc      outa,_c2d
              or        dira,_c2d               '' Enable C2D driver (output)
              nop
              andn      outa,_c2ck              ''strobe 
              or        outa,_c2ck
              nop
              shr       r2,#1 wc
              muxc      outa,_c2d
              nop
              andn      outa,_c2ck              ''strobe 
              or        outa,_c2ck
INS_ret       ret

WAITRDY
              andn      dira,_c2d               '' Disable C2D driver for input
              andn      outa,_c2ck              ''strobe
              or        outa,_c2ck
              test      _c2d,ina wz
        if_z  jmp       #Waitrdy                 '' Strobe C2CK until target transmits a '1' 
WAITRDY_ret   ret


READ8
              andn      dira,_c2d               '' C2D disabled
              mov       r0,#0
              mov       r1,#8
ReadDRlp
              shr       r0,#1
              andn      outa,_c2ck              ''strobe 
              or        outa,_c2ck
              test      _c2d,ina wz
              muxnz     r0,#$80
              djnz      r1,#ReadDRlp
READ8_ret     ret

WRITE8
              or        dira,_c2d
              mov       r1,#8
WRDAT
              shr       r0,#1 wc
              muxc      outa,_c2d      
              andn      outa,_c2ck              ''strobe
              or        outa,_c2ck
              djnz      r1,#WRDAT
WRITE8_ret    ret

start
              andn      dira,_c2d
              andn      outa,_c2ck              '' START - Strobe C2CK with C2D driver disabled
              or        outa,_c2ck
start_ret     ret

stop
              or        outa,_c2d               '' pull it high just for diagnostics
              andn      dira,_c2d               '' STOP
              andn      outa,_c2ck              ''strobe
              or        outa,_c2ck
stop_ret      ret        

'*************************** READ DATA REGISTER ***************************

RD
              call      #start
              mov       r2,#0
              call      #INS
              mov       r2,#0
              call      #INS
              call      #WAITRDY
              call      #READ8
              call      #STOP
RD_ret        ret

'************************* READ ADDRESS REGISTER **************************

RA
              call      #start
              mov       r2,#02
              call      #INS
              call      #Read8
              call      #STOP
RA_ret        ret

'************************** WRITE DATA REGISTER ***************************
              
WD
              call      #start
              mov       r2,#1
              call      #INS
              mov       r2,#0
              call      #INS
              call      #WRITE8
              call      #WAITRDY
              call      #stop
WD_ret        ret

'****************************** WRITE ADDRESS REGISTER *************************
              
WA            call      #start
              mov       r2,#3
              call      #INS
              call      #WRITE8
              call      #stop
WA_ret        ret


_c2ck   long  0         ' set by command to distinguish between target chips
_c2d    long  |<C2D     ' all targets chips use a common data line

zero    long  0


r0      res   1
r1      res   1
r2      res   1
par2    res   1

DAT

{
PPC - Programmable Peripheral Controllers

Many times I have needed dedicated analog inputs for the Propeller with at least 10-bit resolution.
Although there are plenty of A/D converters available that would be suitable it is a different
story when it comes to price. In other projects it has been easier and cheaper for me to use a humble
microcontroller with on-board A/D. The same goes for the Propeller when I need DC accurate A/D.

What I have done is implemented an auto-scanning A/D plus parallel I/O expansion in a cheap micro that can
interface to the Prop in a variety of ways. I2C is the common method as is SPI plus also there are the
serial UART modes.

At present I have this implemented on a Silabs C8051F334 chip which a small 20 pin QFN about the size of a 3mm LED
but I am also looking at implementing this on similar MC9S08 chips that are available in a variety of packages
including DIP.

When I say cheap I mean they are available for a buck or two from Mouser.

Of course this means nothing if you have to write your own software and find a way to program them.
That's why I have already sorted this matter out and the chips can be connected to a Propeller and loaded
from a small Spin loader. I have written firmware in 80C51 assembler for the Silabs part that surpasses
my expectations and the good thing is that it is completely customizable to suit the application.

When I want to upload new firmware I simply select the C2 loader in the Spin tool and hit F10.
The loader runs automatically of course and the firmware files which are embedded in the Spin program
via the FILE directive are loaded into the PPC within a couple of seconds.
After this the loader reboots and you are back to where you were before the interruption as the Prop loads
it's default program from the EEPROM.

I have one system where I have 2 of these chips acting as A/D converters and parallel I/O plus another chip
that runs a front panel with switch inputs and about 40 string of LEDs.
All these chips are connected up to the same I2C lines as the Prop's EEPROM but I have dedicated
an I/O line per chip just for in-circuit reprogramming from the Prop.



PROP PPC (based on C8051F33x)

Interfaces: I2C, SPI, UART
Analog inputs: up to 14 channels x10-bit + temperature
Parallel: up to 14 input/output/quasi  5V tolerant
Peripherals: UART,SPI,I2C,4 Counter/timers

The Silabs parts run at 25MIPS and include a calibrated on-board oscillator +/-2%

Additionally the MC9S08SH4 version includes a low-power RTC.

To talk to my demo PPC chip (as used on CE0810 product).
ADDRESS PPC on I2C
  READING:
    READ 10 bytes as STATUS[1],PORT1[1],ADO[2],AD1[2],AD2[2],AD3[2]
  WRITING:
    COMMAND[1],<DATA>

MY demo PPC COMMANDS:
$00 NOP
$01 WRITE PORT 1 using supplied data and <direction>
    i.e. $01,$F0,$FC writes F0 to port 1 with outputs on 2..7
$02 WRITE PORT 1 DIR REGISTER using supplied data
$FD READ VERSION STRING for up to 32 bytes
$FE READ PROGRAM MEMORY for 32 bytes using supplied address (ADRL,ADRH)
    i.e. $11,$A0,$00 will read the the extended part of the version string at $00A0
$FF RESET PPC

}
DAT
{
PROPOSED PPC FUNCTIONS - based upon MC9S08SH chips

APIO:   Analog and Parallel I/O. Self scanning A/D converter with programmable I/O
        Includes serial line monitoring, Prop reset function and EEPROM autoload
        to support safe EEPROM loading over RS485 and radio networks etc
        Watchdog facility is also supported for resetting the Propeller if it fails
        to write the correct command in time. The watchdog is activated upon receiving
        the "watchdog" command.

LCDKEY: Parallel LCD interface and Keypad scanning. I2C and serial interface
        Supports 4 line big digit mode as well as software controlled contrast which
        can be automatically temperature compensated. The backlight control can also
        be dimmed via a software PWM (External NPN req.)
        
SDCARD: Supports Dual SD card interface using built-in SPI. Access via I2C or serial
        Performs high-level commands thus freeing up a dedicated cog.
        Faster SD access is possible as the PPC's can access the SD CARD using
        dedicated SPI hardware. Serial access up to 1Mbaud is possible. 

EEPROM: EEPROM emulation or multiple Propeller loader. Possibly combined with an
        RTC function including Propeller wakeup via reset.

}